import onnxruntime as onnx
import os
import cupy as np
import yaml
import pandas as pd


# This module is used for predicting and analysis the results.
# Onnxruntime could run on GPU or CPU, which depends on your installed version of this package.
# If you would like to run it on your GPU, uninstall 'onnxruntime' at first, then install 'onnxruntime-gpu'.


class Forward:
    def __init__(self, traits):
        # the path of models, it should be like:
        # ../
        # onnx/
        #   -- model1.onnx
        #   -- modeln.onnx
        # if this path does not exist, it will be automatically create, and weights will be downloaded to it
        if not os.path.exists(r'data/onnx'):
            pass
        self.modelsPath = r'data/onnx/'
        # list of supported Quality Traits and corresponding index to value dictionary
        with open("data/p_trait.yaml") as p:
            self.data_p = yaml.safe_load(p)
        with open("data/n_trait.yaml") as n:
            self.data_n = yaml.safe_load(n)
        # traits to predict
        self.traits = list(set(traits))
        # preload needed models
        self.models = {trait: onnx.InferenceSession(self.modelsPath + trait + '_best.onnx', providers=['CUDAExecutionProvider', 'CPUExecutionProvider']) for trait in self.traits}

    # input is a list of samples
    # here, unpack the input data and split results by different trait types
    def forward(self, index_list, input_data=None, batch_size=1):
        # transform the cupy array to numpy array (to avoid onnxruntime errors)
        input_data = np.asnumpy(input_data)
        # outputs of final results
        df = pd.DataFrame(None, columns=self.traits, index=index_list)
        # predict each trait for each sample
        for trait, model in self.models.items():
            # determine if this trait is a Quality Trait
            if trait in self.data_p.keys():
                # results
                results = []
                # indexed levels dictionary of this trait
                traitLevels = self.data_p[trait]
                # generate batched data and forward them
                for batch in self.batch_generator(input_data, batch_size):

                    # raw results of this batch, its shape is batch * levels' amount
                    out = model.run(['output'], {'input': batch})[0]
                    # turn raw results to indexes
                    out = out.argmax(axis=1)
                    # merge output to list results, for matching levels in only one loop
                    results += out.tolist()
                df[trait] = [traitLevels[result] for result in results]
            # this trait is a Quantity Trait
            else:
                # results
                results = []
                traitMax = self.data_n[trait]['max']
                traitMin = self.data_n[trait]['min']
                traitDiff = traitMax - traitMin
                for batch in self.batch_generator(input_data, batch_size):
                    # raw results of this batch, its shape is batch * levels' amount
                    out = model.run(['output'], {'input': batch})[0]
                    # merge output to list results, for matching levels in only one loop
                    results += out.tolist()
                # denormalize the result
                df[trait] = np.array(results) * traitDiff + traitMin
        return df

    # slice the input data into batches
    def batch_generator(self, input_data, batch_size=1):
        start = 0
        # which batch is running now
        times = 1
        while start < input_data.shape[0]:
            end = start + batch_size
            # load input data batch by batch
            yield input_data[start:end, :, :, :]
            start = end
            times += 1

    def output_dataframe(self, dataframe):
        dataframe.to_excel("./result.xlsx")


import os
import torch


# This module is used for transforming .pt weights to .onnx weights for better performance.
# The inputPath should be like:
# ../
# inputPath/
#   -- weight1.pt
#   -- weightn.pt
# Results will be outputted to inputPath/../onnx

class Transform:
    # transform .pt weights to .onnx weights, input_W and input_H should be given
    def exportAsONNX(self, inputPath, input_W=206, input_H=206, input_channel=3):
        # path of the weights
        inputPath = inputPath
        # loop through files in the path
        for name in os.listdir(inputPath):
            # determine whether it is .pt weight
            if ".pt" in name:
                # print its name
                print(name)
                # load the weight
                model = torch.load(f"{inputPath}/{name}", map_location=torch.device('cuda'))
                # get the input data's channels number
                input_channel = model[0].in_channels
                # generate a random input tensor
                dummy_input = torch.randn(1, input_channel, input_W, input_H).to('cuda')
                # determine whether the output path exists
                if not os.path.exists(f"{inputPath}/../onnx"):
                    # if not exists, create the folder
                    os.mkdir(f"{inputPath}/../onnx")
                # export the model to .onnx, the first size is dynamic, which means it supports batch-input
                dynamic_axes = {'input': {0: 'batch_size'}, 'output': {0: 'batch_size'}}
                torch.onnx.export(model, dummy_input, f"{inputPath}/../onnx/{name.replace('pt', 'onnx')}", input_names=['input'], output_names=['output'], dynamic_axes=dynamic_axes)

# an example
# t = Transform()
# t.exportAsONNX(r"D:\Projects\website\SoybeanWebsite\predict\weight")



import cupy as np
import cudf


# This module is used for read .vcf file, and get filtered SNP data for model input.
# A test .vcf file is provided in example/10_test_examples.vcf. Except contigs, the rest is read as a "\t" delimited csv file.
# The file is like:

# ......
# ##contig=<ID=Chr20>
# ##INFO=<ID=AC,Number=A,Type=Integer,Description="Allele count in genotypes">
# ##INFO=<ID=AN,Number=1,Type=Integer,Description="Total number of alleles in called genotypes">
# ##bcftools_viewVersion=1.8+htslib-1.8
# ##bcftools_viewCommand=view -S test_sample.txt soybean.recode.vcasf.gz; Date=Sun Apr 30 21:11:24 2023
# ##bcftools_viewVersion=1.15.1-26-geafc742+htslib-1.15.1-55-gb7addd3
# ##bcftools_viewCommand=view -S 2.txt test_sample.vcf.gz; Date=Fri May 26 10:42:48 2023
# #CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	PI423734	PI424241	PI437749	PI506711B	PI549076A	PI567453	PI594245B	PI184047	PI229339	PI398202
# Chr01	24952	ss715578788	A	G	.	.	.	GT	0/0	0/0	1/1	1/1	0/0	0/0	0/0	1/1	1/1	0/0
# ......

# After read the file, some transformation steps should be executed to fit the model input.
# In SoyDNGP, a 3*206*206 shaped matrix is needed as input. For detailed explanation, please read README.md.


class Reader:
    # initial class Reader, load the needed SNP file
    def __init__(self, SNPlistPath="data/snp.txt"):
        # open the file and transform it to list
        with open(SNPlistPath) as f:
            # delete '\n' in every elements
            self.SNPlist = [snp.rstrip() for snp in f.readlines()]
        self.columns = []
        self.indexes = None

    # read the file from vcf_path, return a transposed dataframe, whose columns are SNP IDs, and rows are samples' IDs
    def readVCF(self, vcf_path):
        # open the .vcf file
        with open(vcf_path, "r") as f:
            # if the line does not start as "##", this line is the header line
            header = next((i for i, line in enumerate(f) if not line.startswith('##')), 0)
        # read the file from header row
        df = cudf.read_csv(vcf_path, header=header, sep='\t', dtype=str)
        # clean the useless columns, such as ['QUAL', 'FILTER', 'INFO', 'FORMAT', '#CHROM', 'POS', 'REF', 'ALT']
        # after doing that, only ID and sample IDs will be kept
        self.columns.append(df.columns[2])
        self.columns.extend(df.columns[9:])
        # concat the chromosome name and position as new ID for SNPs
        df['ID'] = df['#CHROM'].str.cat(df['POS'], sep='_')
        # set ID column as index and drop the useless columns
        df = df[self.columns].set_index('ID')
        # clear values of DataFrame
        # sometimes, values might be with annotations, for example, '1/1(some annotation)', which should be deleted
        for column in self.columns[1:]:
            # only the first three characters could be kept
            df[column] = df[column].str.slice(0, 3)
            # create a temporary series, which is replaced by rules
            temp_series = df[column].replace({r'1/1': '1', r'1|1': '1', r'0/1': '2', r'0|1': '2'})
            # make a mask to find out unchanged elements
            mask = df[column] == temp_series
            # assign '3' for unchanged elements
            temp_series[mask] = '3'
            # replace column of DataFrame with assigned series
            df[column] = temp_series.astype(np.int32)
            # delete useless variables
            del temp_series, mask
        # transpose the DataFrame, now the SNP IDs are columns, and sample IDs are indexes
        df = df.transpose()
        # keep the indexes and columns
        self.indexes = df.index.values_host
        self.columns = df.columns
        # drop indexes and reduce memory consumption
        df.reset_index(drop=True, inplace=True)
        return df

    # compare processed dataframe's SNP IDs and needed SNP IDs
    def SNPfilter(self, dataframe):
        # find out the missing SNP positions
        missing = list(set(self.SNPlist).difference(set(self.columns)))
        # set a flag to determine if missing is an empty set
        isMissing = bool(missing)
        # determine if there are missing values
        if isMissing:
            # fill the missing values with './.'
            df_missing = cudf.DataFrame(np.full((len(dataframe.index), len(missing)), 3), columns=missing, dtype=np.int32)
            # concat the original dataframe and filling dataframe
            df_filter = cudf.concat([dataframe, df_missing], axis=1)
            # delete useless variable
            del df_missing

        else:
            # if there are not missing values, straightly reset the columns
            df_filter = dataframe[self.SNPlist].astype(np.int32)
        # delete useless variable
        del dataframe, missing
        return df_filter, isMissing

    # convert origin matrix to one-hot matrix, and resize every samples
    def one_hot(self, df_filter):
        # dataframe to numpy array
        matrix = df_filter.values
        # repeat 3 times to 3D, in this way, transforming to one-hot matrix does not need explicit loop, which could perform better
        matrix_3D = np.zeros((matrix.shape[0], matrix.shape[1], 3))
        # the rule for one-hot is like:
        #             '1/1' or '1|1'        '0/1' or '0|1'        other
        # channel 1          1                     1                0
        # channel 2          1                     0                1
        # channel 3          0                     1                1
        # for the first layer, only the first and second condition will be valued as 1
        matrix_3D[:, :, 0] = np.where(np.isin(matrix[:, :], np.array([1, 2])), 1, 0)
        # for the second layer, only the second condition will be valued as 0
        matrix_3D[:, :, 1] = np.where(np.isin(matrix[:, :], np.array(2)), 0, 1)
        # for the third layer, only the first condition will be valued as 0
        matrix_3D[:, :, 2] = np.where(np.isin(matrix[:, :], np.array(1)), 0, 1)
        # resize every sample to shape of model input
        # set an index
        n = -1
        # create batches of resized matrix
        sample_resized = np.empty((matrix.shape[0], 3, 206, 206)).astype(np.float32)
        for sample in matrix_3D:
            # !!!!!(cite from numpy v1.25 document)
            # If the new array is larger than the original array, then the new array is filled with repeated copies of a.
            # Note that this behavior is different from a.resize(new_shape) which fills with zeros instead of repeated copies of a.
            # !!!!!
            sample = np.resize(sample, (206, 206, 3))
            # transpose the order of sample's channels, for np.reshape might change its order
            # for example, before reshape, sample has order [a, b, c], which might be changed after reshape, like [1, c, b, a]
            # to keep the result's order, here a transposition is significant
            sample = np.transpose(sample, (2, 0, 1))
            # the first dimension is batch size, just for fitting the input shape
            sample = np.reshape(sample, (1, 3, 206, 206))
            n += 1
            sample_resized[n, :, :, :] = sample
        return self.indexes, sample_resized

# an example
# r = Reader()  # instantiate class Reader
# print(len(r.SNPlist))  # get length of needed SNP list
# df = r.readVCF(r"D:\Projects\website\SoybeanNext\example\10_test_examples.vcf")  # get the processed dataframe
# df_filter, isMissing = r.SNPfilter(df)  # get the filtered dataframe
# sample_resized, indexlist = r.one_hot(df_filter)  # convert to one-hot matrix and resize every samples




import time
from .reader import Reader
from .forward import Forward
def Run():
    r = Reader()  # instantiate class Reader
    s = time.time()
    df = r.readVCF(r"example/missing_examples.vcf")  # get the processed dataframe
    e = time.time()
    print(f"readVCF————{e - s:.2f}")
    s = time.time()
    df_filter, isMissing = r.SNPfilter(df)  # get the filtered dataframe
    e = time.time()
    print(f"SNPfilter————{e - s:.2f}")
    s = time.time()
    index_list, sample_resized = r.one_hot(df_filter)  # convert to one-hot matrix and resize every samples
    e = time.time()
    print(f"one_hot————{e - s:.2f}")
    s = time.time()
    f = Forward(['MG', 'P_DENS'])  # instantiate class Forward
    df = f.forward(index_list, sample_resized, 30)  # predict and get results
    e = time.time()
    print(f"forward————{e - s: .2f}")
    df.to_csv("./result.csv", index=False)

Run()
